%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% DFBAlab: Dynamic Flux Balance Analysis laboratory                       %
% Process Systems Engineering Laboratory, Cambridge, MA, USA              %
% July 2014                                                               %
% Written by Jose A. Gomez and Kai Hffner                                %
%                                                                         % 
% This code can only be used for academic purposes. When using this code  %
% please cite:                                                            %
%                                                                         %
% Gomez, J.A., Hffner, K. and Barton, P. I.                              %
% DFBAlab: A fast and reliable MATLAB code for Dynamic Flux Balance       %
% Analysis. Submitted.                                                    %
%                                                                         %
% COPYRIGHT (C) 2014 MASSACHUSETTS INSTITUTE OF TECHNOLOGY                %
%                                                                         %
% Read the LICENSE.txt file for more details.                             %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% Syngas bubble column simulation
% Published in BMC Systems Biology
% Written by Jin Chen and Michael A. Henson
% December, 2015

function dy = DRHS(t, y, INFO)

% Assign values

nmodel = INFO.nmodel;
N = INFO.N;
ns = INFO.ns;
condit = INFO.condit;

% Set conditions

klac = condit(1);
Hc = condit(2);
klah = condit(3);
Hh = condit(4);
klac2 = condit(5);
Hc2 = condit(6);
tr = condit(7);
zs = condit(8);
N = condit(9);
ns = condit(10);
ug = condit(11);
ul = condit(12);
dl = condit(13);
eg = condit(14);
el = condit(15);
cgi = condit(16);
hgi = condit(17);
c2gi = condit(18);

% Set other constants

Ma = 60.0/1000;       % Acetate molecular weight g/mmol
Me = 46.1/1000;       % Ethanol molecular weight g/mmol

% Define extracellular state variables

for i=1:N
    Z(i) = y(1+(i-1)*ns);    % biomass concentration, g/L
    cl(i) = y(2+(i-1)*ns);   % liquid CO concentration, mmol/L
    cg(i) = y(3+(i-1)*ns);   % gas CO concentration, mmol/L
    hl(i) = y(4+(i-1)*ns);   % liquid H2 concentration, mmol/L
    hg(i) = y(5+(i-1)*ns);   % gas H2 concentration, mmol/L
    c2l(i) = y(6+(i-1)*ns);  % liquid CO2 concentration, mmol/L
    c2g(i) = y(7+(i-1)*ns);  % gas CO2 concentration, mmol/L
    A(i) = y(8+(i-1)*ns);    % acetate concentration, g/L
    E(i) = y(9+(i-1)*ns);    % ethanol concentration, g/L
end

% The elements of the flux matrix have the sign given to them by the
% coefficients in the Cost vector in main. 
% Example, if:
% C{k}(i).rxns = [144, 832, 931];
% C{k}(i).wts = [3, 1, -1];
% Then the cost vector for this LP will be:
% flux(k,i) = 3*v_144 + v_832 - v_931 
% The penalty is an array containing the feasibility LP problem optimal
% objective function value for each model. 

%% Update bounds and solve for fluxes
[flux,penalty] = solveModel(t,y,INFO);
%%

for i = 1:nmodel
    mu(i) = flux(i,1);      % growth rate, h-1
    vc(i) = flux(i,2);      % CO uptake rate, mmol/g/h
    vh(i) = flux(i,3);      % H2 uptake rate, mmol/g/h
    vc2(i) = flux(i,4);     % CO2 uptake rate, mmol/g/h  
    va(i) = flux(i,5);      % acetate synthesis rate, mmol/g/h 
    ve(i) = flux(i,6);      % ethanol synthesis rate, mmol/g/h
end

%% Dynamics

dy = [];

i=1:N;
    
% Saturation gas concentrations
    
    cls(i) = cg(i)*8.314*tr*Hc/1.013e5*1000;
    hls(i) = hg(i)*8.314*tr*Hh/1.013e5*1000;
    c2ls(i) = c2g(i)*8.314*tr*Hc2/1.013e5*1000;

% Extracellular balances
        
i=1;
        Zi = 1/(ul*zs/(el*dl)+1)*Z(i);
        cli = 1/(ul*zs/(el*dl)+1)*cl(i);
        hli = 1/(ul*zs/(el*dl)+1)*hl(i);  
        c2li = 1/(ul*zs/(el*dl)+1)*c2l(i);
        Ai = 1/(ul*zs/(el*dl)+1)*A(i);
        Ei = 1/(ul*zs/(el*dl)+1)*E(i);
        zd(i) = (Z(i)-Zi)/zs;
        cld(i) = (cl(i)-cli)/zs;
        cgd(i) = (cg(i)-cgi)/zs;
        hld(i) = (hl(i)-hli)/zs;
        hgd(i) = (hg(i)-hgi)/zs;
        c2ld(i) = (c2l(i)-c2li)/zs;
        c2gd(i) = (c2g(i)-c2gi)/zs;
        ad(i) = (A(i)-Ai)/zs;
        ed(i) = (E(i)-Ei)/zs;         
        zd2(i) = (Z(i+1)-2*Z(i)+Zi)/zs^2;
        cld2(i) = (cl(i+1)-2*cl(i)+cli)/zs^2;
        cgd2(i) = (cg(i+1)-2*cg(i)+cgi)/zs^2;        
        hld2(i) = (hl(i+1)-2*hl(i)+hli)/zs^2;
        hgd2(i) = (hg(i+1)-2*hg(i)+hgi)/zs^2;        
        c2ld2(i) = (c2l(i+1)-2*c2l(i)+c2li)/zs^2;
        c2gd2(i) = (c2g(i+1)-2*c2g(i)+c2gi)/zs^2;        
        ad2(i) = (A(i+1)-2*A(i)+Ai)/zs^2;  
        ed2(i) = (E(i+1)-2*E(i)+Ei)/zs^2;
        dZ(i) =  mu(i)*Z(i)-ul*zd(i)/el+dl*zd2(i);
        dcl(i) = vc(i)*Z(i)+klac*(cls(i)-cl(i))/el-ul*cld(i)/el+dl*cld2(i);    
        dcg(i) = -klac*(cls(i)-cl(i))/eg-ug*cgd(i)/eg;   
        dhl(i) = vh(i)*Z(i)+klah*(hls(i)-hl(i))/el-ul*hld(i)/el+dl*hld2(i);     
        dhg(i) = -klah*(hls(i)-hl(i))/eg-ug*hgd(i)/eg;
        dc2l(i) = vc2(i)*Z(i)+klac2*(c2ls(i)-c2l(i))/el-ul*c2ld(i)/el+dl*c2ld2(i);    
        dc2g(i) = -klac2*(c2ls(i)-c2l(i))/eg-ug*c2gd(i)/eg;
        dA(i) =  Ma*va(i)*Z(i)-ul*ad(i)/el+dl*ad2(i);               
        dE(i) =  Me*ve(i)*Z(i)-ul*ed(i)/el+dl*ed2(i); 
        
i=2;
        zd(i) = (2*Z(i+1)+3*Z(i)-6*Z(i-1)+Zi)/(6*zs);
        cld(i) = (2*cl(i+1)+3*cl(i)-6*cl(i-1)+cli)/(6*zs);
        cgd(i) = (2*cg(i+1)+3*cg(i)-6*cg(i-1)+cgi)/(6*zs);
        hld(i) = (2*hl(i+1)+3*hl(i)-6*hl(i-1)+hli)/(6*zs);
        hgd(i) = (2*hg(i+1)+3*hg(i)-6*hg(i-1)+hgi)/(6*zs);
        c2ld(i) = (2*c2l(i+1)+3*c2l(i)-6*c2l(i-1)+c2li)/(6*zs);
        c2gd(i) = (2*c2g(i+1)+3*c2g(i)-6*c2g(i-1)+c2gi)/(6*zs);
        ad(i) = (2*A(i+1)+3*A(i)-6*A(i-1)+Ai)/(6*zs);
        ed(i) = (2*E(i+1)+3*E(i)-6*E(i-1)+Ei)/(6*zs);
        zd2(i) = (Z(i+1)-2*Z(i)+Z(i-1))/zs^2;
        cld2(i) = (cl(i+1)-2*cl(i)+cl(i-1))/zs^2;
        cgd2(i) = (cg(i+1)-2*cg(i)+cg(i-1))/zs^2;
        hld2(i) = (hl(i+1)-2*hl(i)+hl(i-1))/zs^2;
        hgd2(i) = (hg(i+1)-2*hg(i)+hg(i-1))/zs^2;
        c2ld2(i) = (c2l(i+1)-2*c2l(i)+c2l(i-1))/zs^2;
        c2gd2(i) = (c2g(i+1)-2*c2g(i)+c2g(i-1))/zs^2;
        ad2(i) = (A(i+1)-2*A(i)+A(i-1))/zs^2;
        ed2(i) = (E(i+1)-2*E(i)+E(i-1))/zs^2;
        dZ(i) =  mu(i)*Z(i)-ul*zd(i)/el+dl*zd2(i);
        dcl(i) = vc(i)*Z(i)+klac*(cls(i)-cl(i))/el-ul*cld(i)/el+dl*cld2(i);    
        dcg(i) = -klac*(cls(i)-cl(i))/eg-ug*cgd(i)/eg;   
        dhl(i) = vh(i)*Z(i)+klah*(hls(i)-hl(i))/el-ul*hld(i)/el+dl*hld2(i);     
        dhg(i) = -klah*(hls(i)-hl(i))/eg-ug*hgd(i)/eg;
        dc2l(i) = vc2(i)*Z(i)+klac2*(c2ls(i)-c2l(i))/el-ul*c2ld(i)/el+dl*c2ld2(i);    
        dc2g(i) = -klac2*(c2ls(i)-c2l(i))/eg-ug*c2gd(i)/eg;
        dA(i) =  Ma*va(i)*Z(i)-ul*ad(i)/el+dl*ad2(i);               
        dE(i) =  Me*ve(i)*Z(i)-ul*ed(i)/el+dl*ed2(i);

i=3:N-1;
        zd(i) = (2*Z(i+1)+3*Z(i)-6*Z(i-1)+Z(i-2))/(6*zs);
        cld(i) = (2*cl(i+1)+3*cl(i)-6*cl(i-1)+cl(i-2))/(6*zs);
        cgd(i) = (2*cg(i+1)+3*cg(i)-6*cg(i-1)+cg(i-2))/(6*zs);
        hld(i) = (2*hl(i+1)+3*hl(i)-6*hl(i-1)+hl(i-2))/(6*zs);
        hgd(i) = (2*hg(i+1)+3*hg(i)-6*hg(i-1)+hg(i-2))/(6*zs);
        c2ld(i) = (2*c2l(i+1)+3*c2l(i)-6*c2l(i-1)+c2l(i-2))/(6*zs);
        c2gd(i) = (2*c2g(i+1)+3*c2g(i)-6*c2g(i-1)+c2g(i-2))/(6*zs);
        ad(i) = (2*A(i+1)+3*A(i)-6*A(i-1)+A(i-2))/(6*zs);
        ed(i) = (2*E(i+1)+3*E(i)-6*E(i-1)+E(i-2))/(6*zs);
        zd2(i) = (Z(i+1)-2*Z(i)+Z(i-1))/zs^2;
        cld2(i) = (cl(i+1)-2*cl(i)+cl(i-1))/zs^2;
        cgd2(i) = (cg(i+1)-2*cg(i)+cg(i-1))/zs^2;
        hld2(i) = (hl(i+1)-2*hl(i)+hl(i-1))/zs^2;
        hgd2(i) = (hg(i+1)-2*hg(i)+hg(i-1))/zs^2;
        c2ld2(i) = (c2l(i+1)-2*c2l(i)+c2l(i-1))/zs^2;
        c2gd2(i) = (c2g(i+1)-2*c2g(i)+c2g(i-1))/zs^2;
        ad2(i) = (A(i+1)-2*A(i)+A(i-1))/zs^2;
        ed2(i) = (E(i+1)-2*E(i)+E(i-1))/zs^2;
        dZ(i) =  mu(i).*Z(i)-ul*zd(i)/el+dl*zd2(i);
        dcl(i) = vc(i).*Z(i)+klac*(cls(i)-cl(i))/el-ul*cld(i)/el+dl*cld2(i);    
        dcg(i) = -klac*(cls(i)-cl(i))/eg-ug*cgd(i)/eg;   
        dhl(i) = vh(i).*Z(i)+klah*(hls(i)-hl(i))/el-ul*hld(i)/el+dl*hld2(i);     
        dhg(i) = -klah*(hls(i)-hl(i))/eg-ug*hgd(i)/eg;
        dc2l(i) = vc2(i).*Z(i)+klac2*(c2ls(i)-c2l(i))/el-ul*c2ld(i)/el+dl*c2ld2(i);    
        dc2g(i) = -klac2*(c2ls(i)-c2l(i))/eg-ug*c2gd(i)/eg;
        dA(i) =  Ma*va(i).*Z(i)-ul*ad(i)/el+dl*ad2(i);               
        dE(i) =  Me*ve(i).*Z(i)-ul*ed(i)/el+dl*ed2(i);

i=N;

        zd(i) = 0;
        cld(i) = 0;
        cgd(i) = (cg(i)-cg(i-1))/zs;
        hld(i) = 0;
        hgd(i) = (hg(i)-hg(i-1))/zs;
        c2ld(i) = 0;
        c2gd(i) = (c2g(i)-c2g(i-1))/zs;
        ad(i) = 0;
        ed(i) = 0;
        zd2(i) = (2*Z(i-1)-2*Z(i))/zs^2;
        cld2(i) = (2*cl(i-1)-2*cl(i))/zs^2;
        cgd2(i) = (cg(i)-2*cg(i-1)+cg(i-2))/zs^2;
        hld2(i) = (2*hl(i-1)-2*hl(i))/zs^2;
        hgd2(i) = (hg(i)-2*hg(i-1)+hg(i-2))/zs^2;
        c2ld2(i) = (2*c2l(i-1)-2*c2l(i))/zs^2;
        c2gd2(i) = (c2g(i)-2*c2g(i-1)+c2g(i-2))/zs^2;
        ad2(i) = (2*A(i-1)-2*A(i))/zs^2;
        ed2(i) = (2*E(i-1)-2*E(i))/zs^2;
        dZ(i) =  mu(i)*Z(i)-ul*zd(i)/el+dl*zd2(i);
        dcl(i) = vc(i)*Z(i)+klac*(cls(i)-cl(i))/el-ul*cld(i)/el+dl*cld2(i);    
        dcg(i) = -klac*(cls(i)-cl(i))/eg-ug*cgd(i)/eg;   
        dhl(i) = vh(i)*Z(i)+klah*(hls(i)-hl(i))/el-ul*hld(i)/el+dl*hld2(i);     
        dhg(i) = -klah*(hls(i)-hl(i))/eg-ug*hgd(i)/eg;
        dc2l(i) = vc2(i)*Z(i)+klac2*(c2ls(i)-c2l(i))/el-ul*c2ld(i)/el+dl*c2ld2(i);    
        dc2g(i) = -klac2*(c2ls(i)-c2l(i))/eg-ug*c2gd(i)/eg;
        dA(i) =  Ma*va(i)*Z(i)-ul*ad(i)/el+dl*ad2(i);               
        dE(i) =  Me*ve(i)*Z(i)-ul*ed(i)/el+dl*ed2(i); 

i=1:N;
dyTmp = [dZ(i); dcl(i); dcg(i); dhl(i); dhg(i); dc2l(i); dc2g(i); dA(i); dE(i)];
dyTmp1 = reshape(dyTmp,[],1);
dy=[dy dyTmp1'];
        

dylast=0;
for i=1:nmodel
    dylast = dylast + penalty(i);
end
dy = [dy dylast];
dy = dy';
end