%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% DFBAlab: Dynamic Flux Balance Analysis laboratory                       %
% Process Systems Engineering Laboratory, Cambridge, MA, USA              %
% July 2014                                                               %
% Written by Jose A. Gomez and Kai Hffner                                %
%                                                                         % 
% This code can only be used for academic purposes. When using this code  %
% please cite:                                                            %
%                                                                         %
% Gomez, J.A., Hffner, K. and Barton, P. I.                              %
% DFBAlab: A fast and reliable MATLAB code for Dynamic Flux Balance       %
% Analysis. Submitted.                                                    %
%                                                                         %
% COPYRIGHT (C) 2014 MASSACHUSETTS INSTITUTE OF TECHNOLOGY                %
%                                                                         %
% Read the LICENSE.txt file for more details.                             %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% Syngas bubble column simulation - main program
% Published in BMC Systems Biology
% Written by Jin Chen and Michael A. Henson
% December, 2015

clear all

% Set number of models and simulation time
nmodel = 100;
tspan = [0,250];
N = nmodel;                     % number of reactor discretization points
ns = 9;                         % number of state variables

% Load models. These should be .mat files generated by the COBRA toolbox. 
% When generating these files using the COBRA toolbox, a big number is used
% as infinity. This number should be fed to the DB vector (Default bound).
% INFO.DB = DB;

for i = 1:nmodel
    load model.mat 
    model{i} = CL_Auto; 
    DB(i) = 1000;
end

%% exID array
% You can either search the reaction names by name or provide them directly
% in the exID array.
% RxnNames = {'EX_glc(e)', 'EX_ac(e)', 'biomass'};
% for i = 1:length(RxnNames)
%    [a,exID(i)] = ismember(RxnNames(i),model.rxns);
% end
% INFO.exID = exID;
% Lower bounds and upper bounds for these reactions should be provided in
% the RHS code. 

for i = 1:nmodel
    exID{i}=[20, 21, 43, 1, 26, 4];
end

% This codes solves the LPs in standard form. Bounds on exchange fluxes in 
% the exID array can be modified directly on the first 2*n rows where n is 
% the number of exchange fluxes. Order will be lower bound, upper bound, 
% lower bound, upper bound in the same order as exID. 
%
% NOTE: All bounds on fluxes in the exID arrays are relaxed to -Inf and 
% + Inf. These bounds need to be updated if needed in the RHS file.

%% Cost vectors
% Usually the first cost vector will be biomass maximization, but it can
% be any other objective. The CPLEX objects will minimize by default. 
% Report only nonzero elements. 
% The structure should be:
% C{model} = struct
% Element C{k}(i) of C, is the cost structure i for model k. 
% C{k}(i).sense = +1 for minimize, or -1 for maximize.
% C{k}(i).rxns = array containing the reactions in this objective. 
% C{k}(i).wts = array containing coefficients for reactions reported in 
% rxns. Both arrays should have the same length. 
% Example, if:
% C{k}(i).rxns = [144, 832, 931];
% C{k}(i).wts = [3, 1, -1];
% Then the cost vector for this LP will be:
% Cost{k}(i) = 3*v_144 + v_832 - v_931 (fluxes for model k). 
% This cost vector will be either maximized or minimized depending on the
% value of C{k}(i).sense.

% In SBML files, usually production fluxes are positive and uptake fluxes
% are negative. Keep in mind that maximizing a negative flux implies 
% minimizing its absolute value.
% Different models can have different number of objectives. 
% INFO.C = C; 

minim = 1;
maxim = -1;

for i=1:nmodel
    
    % Maximize growth
    C{i}(1).sense = maxim;
    C{i}(1).rxns = [1];
    C{i}(1).wts = [1];
    % Maximize CO uptake
    C{i}(2).sense = minim;
    C{i}(2).rxns = [20];
    C{i}(2).wts = [1];
    % Maximize H2 uptake
    C{i}(3).sense = minim;
    C{i}(3).rxns = [43];
    C{i}(3).wts = [1];
    % Minimize CO2 production
    C{i}(4).sense = minim;
    C{i}(4).rxns = [21];
    C{i}(4).wts = [1];
    % Minimize acetate production
    C{i}(5).sense = minim;
    C{i}(5).rxns = [4];
    C{i}(5).wts = [1];
    % Minimize ethanol production
    C{i}(6).sense = minim;
    C{i}(6).rxns = [26];
    C{i}(6).wts = [1];

end

%%
% Pass parameters

INFO.nmodel = nmodel;
INFO.DB = DB;
INFO.exID = exID;
INFO.C = C; 

% Set operating conditions

ug = 75;                        % constant gas velocity, m/h
L = 25;                         % reactor length, m 
A = 5;                          % reactor cross-sectional area, m^2
zs = L/N;                       % spatial step size, m
ul = 0.25;                      % inlet liquid velocity, m/h
dl = 0.25;                      % liquid dispersion coefficient, m^2/h
tr = 310.15;                    % temperature, C
pL = 1.013e5;                   % pressure at top of column, Pa
po = pL+1000*9.81*L;            % pressure at bottom of column, Pa
pc = 0.6;                       % partial pressure of CO %(60+2*(mm-1))/100;
Hc = 8.0e-4;                    % Henry's constant for CO in water, mol/L*atm
ph = 1-pc;                      % partial pressure of H2
Hh = 6.6e-4;                    % Henry's constant for H2 in water, mol/L*atm
pc2 = 0;                        % partial pressure of CO2
Hc2 = 2.5e-2;                   % Henry's constant for CO2 in water, mol/L*atm
cgi = pc*po/8.314/tr;           % CO feed concentration, mol/m3 = mmol/L
hgi = ph*po/8.314/tr;           % H2 feed concentration, mol/m3 = mmol/L
c2gi = pc2*po/8.314/tr;         % CO2 feed concentration, mol/m3 = mmol/L

% Set mass transfer coefficients
klac = 80;                      % CO gas-liquid mass transfer coefficient, h-1  
klah = 2.5*klac;                % H2 gas-liquid mass transfer coefficient, h-1
klac2 = klac;                   % CO2 gas-liquid mass transfer coefficient, h-1

% Calculate gas and liquid holdups

ugms=ug/3600;
eg = 0.53*ugms/(0.15+ugms);
el = 1-eg;

% Form condition vector

condit = [klac,Hc,klah,Hh,klac2,Hc2,tr,zs,N,ns,ug,ul,dl,eg,el,cgi,hgi,c2gi];

% Set uptake parameters

vcm = 35;
vhm = 70;
Kmc = 0.02;
Kmh = 0.02;
Kie = 10;
param = [
     vcm                 % maximum CO uptake rate, mmol/g/h; 34.36 in Mohammodi 2014
     Kmc                 % CO uptake saturation constant, mmol/L; 0.021atm in Mohamodi 2014 
     vhm                 % maximum H2 uptake rate, mmol/g/h 
     Kmh                 % H2 uptake saturation constant, mmol/L;
     vcm                 % maximum CO2 uptake rate, mmol/g/h 
     Kmc                 % CO2 uptake saturation constant, mmol/L;
     Kie                 % product inhibition constant, mmol/L
    ];

% Pass parameters

INFO.param = param;
INFO.ns = ns;
INFO.N = N;
INFO.condit = condit;

% Set initial conditions

cls = pc*po*Hc*1000/1.013e5;
hls = ph*po*Hh*1000/1.013e5;
c2ls = pc2*po*Hc2*1000/1.013e5;
yo = [];
yz = [0.1 cls cgi hls hgi c2ls c2gi 0 0];

for i = 1:N
    yo = [yo yz];
end
yo = [yo 0];
Y0 = yo;

% CPLEX Objects construction parameters

INFO.LPsolver = 1; % CPLEX = 0, Gurobi = 1.
                   % CPLEX works equally fine with both methods.
                   % Gurobi seems to work better with Method = 1, and 
                   % Mosek with Method = 0.
INFO.tol = 1E-9; % Feasibility, optimality and convergence tolerance for Cplex (tol>=1E-9). 
                 % It is recommended it is at least 2 orders of magnitude
                 % tighter than the integrator tolerance. 
                 % If problems with infeasibility messages, tighten this
                 % tolerance.
INFO.tolPh1 = INFO.tol; % Tolerance to determine if a solution to phaseI equals zero.
                   % It is recommended to be the same as INFO.tol. 
INFO.tolevt = 10*INFO.tol; % Tolerance for event detection. Has to be greater 
                   % than INFO.tol.

% You can modify the integration tolerances here.
% If some of the flows become negative after running the simulation once
% you can add the 'Nonnegative' option.

options = odeset('AbsTol',1E-6,'RelTol',1E-6, ...
'NonNegative',1:N*ns,'Events',@evts, ...
'OutputSel',1,'OutputFcn',@odeplot, ...
'JPattern',syngas_bubblecolumn5_jac(N-1,ns));

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
[model,INFO] = ModelSetupM(model,Y0,INFO);
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

if INFO.LPsolver == 0
    [INFO] = LexicographicOpt(model,INFO);
elseif INFO.LPsolver == 1
    [INFO] = LexicographicOptG(model,INFO);
else
    display('Solver not currently supported.');
end

tint = 0;
TF = [];
YF = [];
figure(1000)
while tint<tspan(2)
    
% Look at MATLAB documentation if you want to change solver.
% ode15s is more or less accurate for stiff problems. 
%     [T,Y] = ode15s(@DRHS,tspan,Y0,options,INFO);

    [T,Y] = ode15s(@DRHS,tspan,Y0,options,INFO);
    TF = [TF;T];
    YF = [YF;Y];
    tint = T(end);
    tspan = [tint,tspan(2)];
    Y0 = Y(end,:);
    if tint == tspan(2)
        break;
    end
    
% Determine model with basis change

    value = evts(tint,Y0,INFO);
    [jjj,j] = min(value);
    ct = 0;
    k = 0;
    while j>ct
        k = k + 1;
        ct = ct + size(model{k}.A,1);
    end
    INFO.flagbasis = k;
    fprintf('Basis change at time %d. \n',tint);
    
% Update b vector 

[INFO] = bupdate(tint,Y0,INFO);

% Perform lexicographic optimization

if INFO.LPsolver == 0
    [INFO] = LexicographicOpt(model,INFO);
elseif INFO.LPsolver == 1
    [INFO] = LexicographicOptG(model,INFO);
else
    display('Solver not currently supported.');
end

end

T = TF;
Y = YF;
     
%% Plotting

figure(1)
mz = round(N/2)-1;
subplot(3,3,1)
hold on
plot(T,Y(:,ns*mz+1))
hold on
plot(T,Y(:,ns*mz+8))
hold on
plot(T,Y(:,ns*mz+9))
xlabel('Time [h]','FontSize',8)
ylabel('Biomass, Acetate, Ethanol [g/L]','FontSize',8)
legend({'Biomass', 'Acetate', 'Ethanol'},'Location','best','FontSize',8)
title('Middle Column','FontSize',8)

subplot(3,3,4)
hold on
plot(T,Y(:,ns*mz+2))
hold on
plot(T,Y(:,ns*mz+4))
hold on
plot(T,Y(:,ns*mz+6))
xlabel('Time [h]','FontSize',8)
ylabel('CO, H2, CO2 in liquid [mmol/L]','FontSize',8)
legend({'CO','H2','CO2'},'Location','best','FontSize',8)

subplot(3,3,7)
hold on
plot(T,Y(:,ns*mz+3))
hold on
plot(T,Y(:,ns*mz+5))
hold on
plot(T,Y(:,ns*mz+7))
xlabel('Time [h]','FontSize',8)
ylabel('CO, H2, CO2 in gas [mmol/L]','FontSize',8)
legend({'CO','H2','CO2'},'Location','best','FontSize',8)

mz=N-1;

subplot(3,3,2)
hold on
plot(T,Y(:,ns*mz+1))
hold on
plot(T,Y(:,ns*mz+8))
hold on
plot(T,Y(:,ns*mz+9))
xlabel('Time [h]','FontSize',8)
ylabel('Biomass, Acetate, Ethanol [g/L]','FontSize',8)
legend({'Biomass', 'Acetate', 'Ethanol'},'Location','best','FontSize',8)
title('Top Column','FontSize',8)

subplot(3,3,5)
hold on
plot(T,Y(:,ns*mz+2))
hold on
plot(T,Y(:,ns*mz+4))
hold on
plot(T,Y(:,ns*mz+6))
xlabel('Time [h]','FontSize',8)
ylabel('CO, H2, CO2 in liquid [mmol/L]','FontSize',8)
legend({'CO','H2','CO2'},'Location','best','FontSize',8)

subplot(3,3,8)
hold on
plot(T,Y(:,ns*mz+3))
hold on
plot(T,Y(:,ns*mz+5))
hold on
plot(T,Y(:,ns*mz+7))
xlabel('Time [h]','FontSize',8)
ylabel('CO, H2, CO2 in gas [mmol/L]','FontSize',8)
legend({'CO','H2','CO2'},'Location','best','FontSize',8)

% Calculate spatial profiles at final time

for i=1:nmodel
    zt(i)=i*zs;
    yfin1(i)=Y(end,ns*(i-1)+1);
    yfin2(i)=Y(end,ns*(i-1)+2);
    yfin3(i)=Y(end,ns*(i-1)+3);
    yfin4(i)=Y(end,ns*(i-1)+4);
    yfin5(i)=Y(end,ns*(i-1)+5);
    yfin6(i)=Y(end,ns*(i-1)+6);
    yfin7(i)=Y(end,ns*(i-1)+7);
    yfin8(i)=Y(end,ns*(i-1)+8);
    yfin9(i)=Y(end,ns*(i-1)+9);
end

subplot(3,3,3)
hold on
plot(zt,yfin1)
hold on
plot(zt,yfin8)
hold on
plot(zt,yfin9)
xlabel('Location [m]','FontSize',8)
ylabel('Biomass, Acetate, Ethanol [g/L]','FontSize',8)
legend({'Biomass', 'Acetate', 'Ethanol'},'Location','best','FontSize',8)
xlim([0 L])
title('Spatial profile at Steady State','FontSize',8)

subplot(3,3,6)
hold on
plot(zt,yfin2)
hold on
plot(zt,yfin4)
hold on
plot(zt,yfin6)
xlabel('Location [m]','FontSize',8)
ylabel('CO, H2, CO2 in liquid [mmol/L]','FontSize',8)
legend({'CO','H2','CO2'},'Location','best','FontSize',8)
xlim([0 L])

subplot(3,3,9)
hold on
plot(zt,yfin3)
hold on
plot(zt,yfin5)
hold on
plot(zt,yfin7)
xlabel('Location [m]','FontSize',8)
ylabel('CO, H2, CO2 in gas [mmol/L]','FontSize',8)
legend({'CO','H2','CO2'},'Location','best','FontSize',8)
xlim([0 L])